#ifndef _SEGWAYRMP400_H
#define _SEGWAYRMP400_H

#include <segway_rmp200.h>
#include <ftdiserver.h>
#include <vector>
#include <string>
#include <iostream>

#define RMP400_b 0.54 // distance between weels for one rmp unit in meters
#define RMP200_DEFAULT_WHEEL_RADIUS 0.24

const unsigned int NUM_SEGWAY_200 = 2;

enum SegwayRMP400_status { rmp400_off, rmp400_connected };

struct TSegwayRMP400Status
{
    TSegwayRMP200Status rmp200[2];
};

class CSegwayRMP400
{
  private:
    std::vector<CSegwayRMP200 *> segways_;
    /* vector of valid segway serials found in the ftdi bus */
    std::vector<std::string> serial_ftdi_segway_devices_;

    /**
     * \brief Segway status
     */
    SegwayRMP400_status status_;

    /**
     * \brief forward displacement
     *
     * This value has the displacement of the whole platform in meters
     * TODO: should be a thread periodically updating this value
     */
    float forward_displacement_;

    /**
     * \brief yaw displacement
     *
     * This value has the rotation of the whole platform in revolutions.
     * TODO: should be a thread periodically updating this value
     */
    float yaw_displacement_;

    /**
     * \brief yaw rate
     *
     * This value has the yaw rate of the platform in degrees per second.
     * TODO: should be a thread periodically updating this value
     */
    float yaw_rate_;

    /**
     * \brief default rmp400 wheel radius when inflated to 10 psi (in meters)
     */
    static const float default_wheel_radius_ = 0.26;

    /**
     * \brief wheel radius in meters
     *
     */
    float wheel_radius_;

    /**
     * \brief whether use or not the correction factor to fix wheel size
     * (default is true). See fix_with_wheel_factor for more info.
     *
     */
    bool use_correction_wheel_factor_;

    /**
     * \brief apply a corretion factor to the value given
     *
     * It is suspected that segway is using the RMP200 wheel radius in RMP400
     * firmware which gives a bad measure for displacements. This functions
     * calculate the relation between radius and fix the value given using this
     * factor.
     */
    void fix_with_wheel_factor(float & value);

    /**
     * \brief disable use of correction wheel factor for displacements
     *
     * Check fix_with_wheel_factor for more info.
     */
    void disable_corretion_wheel_factor();

    void scan_ftdi_for_segways();

    /**
     * \brief set a different wheel radius (in meters)
     *
     * Default, when wheels are inflated to 10psi, is 0.26m. If the segway has
     * a different meassure, please use this function before starting to operate
     * as normal.
     */
    void set_wheel_radius(const float radius);

    /*
     * \brief check if the driver is already connected (made a start() call).
     *
     * It will return an RMP400NoConnected exception if fail
     */
    void need_to_be_connected() const;

    /* \brief aux function to check if platform is really doing the reset in
     *        integrators.
     */
    bool is_valid_reset_integrators();

  protected:
    /**
    * \brief allowed modes:
    * - tractor
    * - power_dow
    */
    op_mode mode;

  public:
    CSegwayRMP400();
    CSegwayRMP400(const std::string serial1, const std::string serial2);

    ~CSegwayRMP400();

    /**
     * \brief function to set the operation mode
     *
     * Operation modes: tractor or power_down.
     * In the RMP400 the balance mode is disabled.
     */
    void set_operation_mode(op_mode mode);

    /**
     * \brief function to check if segway is already connected
     */
    bool is_connected() const;

    /**
     * \brief function to return whole platform status
     *
     * It will return the two segway internal status
     */
    TSegwayRMP400Status get_status(void);

    /**
     * \brief function to return the total forward displacement
     *
     * This function returns the current forward displacement in meters. This
     * function only returns the value of the internal attribute, but it does
     * not access the hardware platform. This value is periodically updated by
     * the feedback thread.
     *
     * Both fordward displacement from segways are sum and a mean is calculated
     *
     * \return the current forward displacement in meters.
     *
     */
    float get_forward_displacement(void);

    /**
     * \brief function to return the total yaw displacement
     *
     * This function returns the current yaw displacement in revolutions per
     * second. This function only returns the value of the internal attribute,
     * but it does not access the hardware platform. This value is periodically
     * updated by the feedback thread.
     *
     * Both yaw displacement from segways are sum and a mean is calculated
     *
     * \return the current yaw displacement in revolutions per second.
     *
     */
    float get_yaw_displacement(void);

    /**
     * \brief function to return the yaw rate
     *
     * This function returns the current yaw rate in revolutions per second.
     * This function only returns the value of the internal attribute, but it
     * does not access the hardware platform. This value is periodically updated
     * by the feedback thread.
     *
     * Both yaw rate from segways are sum and a mean is calculated
     *
     * \return the current yaw rate in revolutions per second.
     *
     */
    float get_yaw_rate(void);

    /**
    * \brief function to reset integrators
    *
    * This function sets all encoder integrators to zero in both platforms.
    * Currently this function seems NOT to be reliable since hardware not
    * always is doing the proper work.
    */
    void reset_integrators();

    /**
    * \brief function to move the platform
    *
    * This function sets the rotational and translational velocities to a desired value
    * vT is in m/sec, vR is in rev/sec
    *
    * Rotation about the center of the platform with skid steer is disabled.
    * Skid steering is not recommended since it produces significant tire wear and unnecessary battery consumption.
    * Odometry readings during skid steering are useless since they depend heavily on soil characteristics.
    * A minimum Instantaneous Center of Rotation (ICR) located on the side of the platform is enforced (0.54cm).
    */
    void move(float vT, float vR);

    /* \brief function to connect segways RMP200
     *
     * This functions start the vehicle
     */
    void start();

    /* \brief function to stop the platform
    *
    * This function stops the vehicle
    */
    void stop();

    void close();

  friend std::ostream& operator<< (std::ostream& out, CSegwayRMP400& segway);
};

inline bool
CSegwayRMP400::is_connected() const
{
    return (status_ == rmp400_connected);
}


#endif


